@@ -24,6 +24,7 @@ static s6_servicedir_desc const svdir_file_list[] =
   { .name = "timeout-finish", .type = S6_FILETYPE_UINT, .options = 0 },
   { .name = "max-death-tally", .type = S6_FILETYPE_UINT, .options = 0 },
   { .name = "down-signal", .type = S6_FILETYPE_NORMAL, .options = 0 },
+  { .name = "flag-newpidns", .type = S6_FILETYPE_EMPTY, .options = 0 },
   { .name = "template", .type = S6_FILETYPE_DIR, .options = 0 },
   { .name = "data", .type = S6_FILETYPE_DIR, .options = 0 },
   { .name = "env", .type = S6_FILETYPE_DIR, .options = 0 },
@@ -83,6 +84,11 @@ int s6rc_servicedir_copy_one (char const *src, char const *dst, s6_servicedir_de
       {
         if (errno != ENOENT || p->options & S6_SVFILE_MANDATORY) return 0 ;
       }
+      if (!strcmp(p->name, "template"))
+      {
+        memcpy(dstfn + dstlen + 1, "instance", 9) ;
+        if (mkdir(dstfn, 0755) == -1 && errno != EEXIST) return 0 ;
+      }
       break ;
     default : return (errno = EDOM, 0) ;
   }
