@@ -8,18 +8,8 @@
 #include <errno.h>
 #include <stdlib.h>  /* qsort */
 
-#include <skalibs/types.h>
-#include <skalibs/bitarray.h>
-#include <skalibs/strerr2.h>
-#include <skalibs/sgetopt.h>
-#include <skalibs/allreadwrite.h>
-#include <skalibs/buffer.h>
-#include <skalibs/cdbmake.h>
-#include <skalibs/direntry.h>
-#include <skalibs/djbunix.h>
-#include <skalibs/stralloc.h>
-#include <skalibs/genalloc.h>
-#include <skalibs/skamisc.h>
+#include <skalibs/envexec.h>
+#include <skalibs/stddjb.h>
 #include <skalibs/avltree.h>
 #include <skalibs/unix-transactional.h>
 
@@ -37,9 +27,17 @@
 
 #define S6RC_INTERNALS "s6-rc-compile internals"
 
-static unsigned int verbosity = 1 ;
-static stralloc keep = STRALLOC_ZERO ;
-static stralloc data = STRALLOC_ZERO ;
+enum golb_e
+{
+  GOLB_BLOCK = 0x01,
+} ;
+
+enum gola_e
+{
+  GOLA_VERBOSITY,
+  GOLA_FDHUSER,
+  GOLA_N
+} ;
 
 typedef enum servicetype_e servicetype_t, *servicetype_t_ref ;
 enum servicetype_e
@@ -50,6 +48,10 @@ enum servicetype_e
   SVTYPE_BUNDLE = 4
 } ;
 
+static unsigned int verbosity = 1 ;
+static stralloc keep = STRALLOC_ZERO ;
+static stralloc data = STRALLOC_ZERO ;
+
 
  /* The names tree */
 
@@ -145,6 +147,21 @@ static char const *typestr (servicetype_t type)
          "unknown" ;
 }
 
+static inline int type_check (int dfd)
+{
+  char type[8] ;
+  ssize_t r = openreadnclose_at(dfd, "type", type, 8) ;
+  if (r == -1) return -1 ;
+  if (r < 6) return 0 ;
+  if (type[r-1] == '\n') r-- ;
+  if (r == 8) return 0 ;
+  type[r++] = 0 ;
+  return
+    !strcmp(type, "longrun") ? 1 :
+    !strcmp(type, "oneshot") ? 2 :
+    !strcmp(type, "bundle")  ? 3 : 0 ;
+}
+
 static int add_name_nocheck (before_t *be, char const *srcdir, char const *name, servicetype_t type, unsigned int *pos, unsigned int *kpos)
 {
   uint32_t id ;
@@ -239,7 +256,7 @@ static unsigned int add_internal_longrun (before_t *be, char const *name)
     {
       .ndeps = 0,
       .depindex = genalloc_len(unsigned int, &be->indices),
-      .annotation_flags = 0,
+      .annotation_flags = S6RC_DB_FLAG_ESSENTIAL,
       .timeout = { 0, 0 }
     },
     .srcdir = 0,
@@ -366,8 +383,8 @@ static uint32_t read_timeout (int dfd, char const *srcdir, char const *name, cha
 {
   char buf[64] ;
   uint32_t timeout = 0 ;
-  size_t r = openreadnclose_at(dfd, tname, buf, 63) ;
-  if (!r)
+  ssize_t r = openreadnclose_at(dfd, tname, buf, 63) ;
+  if (r == -1)
   {
     if (errno && errno != ENOENT)
       strerr_diefu6sys(111, "read ", srcdir, "/", name, "/", tname) ;
@@ -387,6 +404,7 @@ static uint32_t read_flags (int dfd, char const *srcdir, char const *name)
   static char const *files[32] =
   {
     "flag-essential",
+    "flag-recommended",
     0
   } ;
   uint32_t flags = 0 ;
@@ -527,26 +545,15 @@ static inline void add_bundle (before_t *be, int dfd, char const *srcdir, char c
 
 static inline void add_source (before_t *be, int dfd, char const *srcdir, char const *name)
 {
-  char typestr[8] = "" ;
-  size_t r ;
-  if (verbosity >= 2) strerr_warni4x("parsing ", srcdir, "/", name) ;
-  r = openreadnclose_at(dfd, "type", typestr, 8) ;
-  if (!r)
+  if (verbosity >= 3) strerr_warni4x("parsing ", srcdir, "/", name) ;
+  switch (type_check(dfd))
   {
-    if (!errno) errno = EINVAL ;
-    strerr_diefu5sys(111, "read ", srcdir, "/", name, "/type") ;
+    case 0 : strerr_dief6x(1, "invalid ", srcdir, "/", name, "/type", ": must be oneshot, longrun, or bundle") ;
+    case 1 : add_longrun(be, dfd, srcdir, name) ; break ;
+    case 2 : add_oneshot(be, dfd, srcdir, name) ; break ;
+    case 3 : add_bundle(be, dfd, srcdir, name) ; break ;
+    default : strerr_diefu5sys(111, "read ", srcdir, "/", name, "/type") ;
   }
-  if (typestr[r-1] == '\n') r-- ;
-  typestr[r++] = 0 ;
-  if (!strcmp(typestr, "oneshot")) add_oneshot(be, dfd, srcdir, name) ;
-  else if (!strcmp(typestr, "longrun")) add_longrun(be, dfd, srcdir, name) ;
-  else if (!strcmp(typestr, "bundle")) add_bundle(be, dfd, srcdir, name) ;
-  else strerr_dief6x(1, "invalid ", srcdir, "/", name, "/type", ": must be oneshot, longrun, or bundle") ;
-}
-
-static int qsort_cannot_use_strcmp_directly (void const *a, void const *b)
-{
-  return strcmp(*(char const *const *)a, *(char const *const *)b) ;
 }
 
 static inline void add_sources (before_t *be, char const *srcdir, stralloc *sa)
@@ -555,6 +562,7 @@ static inline void add_sources (before_t *be, char const *srcdir, stralloc *sa)
   int fdsrc ;
   DIR *dir = opendir(srcdir) ;
   if (!dir) strerr_diefu2sys(111, "opendir ", srcdir) ;
+  if (verbosity >= 2) strerr_warni("adding sources from ", srcdir) ;
   fdsrc = dir_fd(dir) ;
   for (;;)
   {
@@ -583,7 +591,7 @@ static inline void add_sources (before_t *be, char const *srcdir, stralloc *sa)
       names[i] = sa->s + pos ;
       pos += strlen(sa->s + pos) + 1 ;
     }
-    qsort(names, n, sizeof(char const *), &qsort_cannot_use_strcmp_directly) ;
+    qsort(names, n, sizeof(char const *), &str_cmp) ;
     for (unsigned int i = 0 ; i < n ; i++)
     {
       int fd = open_readatb(fdsrc, names[i]) ;
@@ -1087,6 +1095,7 @@ static inline void write_oneshot_runner (char const *compiled, int blocking)
 
 static inline void write_fdholder (char const *compiled, s6rc_db_t const *db, char const *fdhuser)
 {
+  unsigned int nfds = 0 ;
   size_t base = satmp.len ;
   char fn[61 + S6RC_FDHOLDER_LEN] = "servicedirs/" S6RC_FDHOLDER "/data/rules/uid/0/env" ;
   make_skel(compiled, S6RC_FDHOLDER, 1) ;
@@ -1106,8 +1115,12 @@ static inline void write_fdholder (char const *compiled, s6rc_db_t const *db, ch
 
   for (uint32_t j = 0 ; j < db->nlong ; j++)
     if (db->services[j].x.longrun.nproducers)
+    {
       if (!stralloc_cats(&satmp, db->string + db->services[j].name)
        || !stralloc_catb(&satmp, "\n", 1)) dienomem() ;
+      nfds += 2 ;
+    }
+  if (nfds < 240) nfds = 256 ; else nfds += (nfds >> 4) + 2 ;
 
   auto_file(compiled, "servicedirs/" S6RC_FDHOLDER "/data/autofilled", satmp.s + base, satmp.len - base) ;
   satmp.len = base ;
@@ -1129,7 +1142,13 @@ static inline void write_fdholder (char const *compiled, s6rc_db_t const *db, ch
      || !string_quote(&satmp, fdhuser, strlen(fdhuser))
      || !stralloc_catb(&satmp, "\n", 1)) dienomem() ;
   }
-  if (!stralloc_cats(&satmp, S6_EXTBINPREFIX "s6-fdholder-daemon -1 ")) dienomem() ;
+  if (!stralloc_cats(&satmp, S6_EXTBINPREFIX "s6-fdholder-daemon -1 -n ")) dienomem() ;
+  {
+    char fmt[UINT_FMT] ;
+    size_t len = uint_fmt(fmt, nfds) ;
+    fmt[len++] = ' ' ;
+    if (!stralloc_catb(&satmp, fmt, len)) dienomem() ;
+  }
   if (fdhuser)
   {
     if (!stralloc_cats(&satmp, "-U ")) dienomem() ;
@@ -1189,7 +1208,7 @@ static inline void write_resolve (char const *compiled, s6rc_db_t const *db, bun
     }
   }
 
-  if (!cdbmake_finish(&c) || fsync(fd) < 0)
+  if (!cdbmake_finish(&c))
   {
     cleanup(compiled) ;
     strerr_diefu2sys(111, "write to ", fn) ;
@@ -1197,12 +1216,12 @@ static inline void write_resolve (char const *compiled, s6rc_db_t const *db, bun
   close(fd) ;
 }
 
-static void dircopy (char const *compiled, char const *srcfn, char const *dstfn)
+static int dircopy (char const *compiled, char const *srcfn, char const *dstfn)
 {
   struct stat st ;
   if (stat(srcfn, &st) < 0)
   {
-    if (errno == ENOENT) return ;
+    if (errno == ENOENT) return 0 ;
     cleanup(compiled) ;
     strerr_diefu2sys(111, "stat ", srcfn) ;
   }
@@ -1216,13 +1235,13 @@ static void dircopy (char const *compiled, char const *srcfn, char const *dstfn)
     cleanup(compiled) ;
     strerr_diefu4sys(111, "recursively copy ", srcfn, " to ", dstfn) ;
   }
+  return 1 ;
 }
 
-static void write_exe_wrapper (char const *compiled, char const *fn, s6rc_db_t const *db, unsigned int i, char const *exe, int needargs)
+static void write_exe_wrapper (char const *compiled, char const *fn, s6rc_db_t const *db, unsigned int i, char const *exe)
 {
   size_t base = satmp.len ;
-  if (!stralloc_cats(&satmp, "#!" EXECLINE_SHEBANGPREFIX "execlineb -")
-   || !stralloc_cats(&satmp, needargs ? "S0\n" : "P\n")) dienomem() ;
+  if (!stralloc_cats(&satmp, "#!" EXECLINE_SHEBANGPREFIX "execlineb -S0\n")) dienomem() ;
   if (db->services[i].x.longrun.nproducers)
   {
     if (!stralloc_cats(&satmp, S6_EXTBINPREFIX "s6-fdholder-retrieve ../s6rc-fdholder/s \"pipe:s6rc-r-")
@@ -1240,8 +1259,7 @@ static void write_exe_wrapper (char const *compiled, char const *fn, s6rc_db_t c
   }
   if (!stralloc_cats(&satmp, "./")
    || !stralloc_cats(&satmp, exe)
-   || !stralloc_cats(&satmp, ".user")
-   || !stralloc_cats(&satmp, needargs ? " $@\n" : "\n")) dienomem() ;
+   || !stralloc_cats(&satmp, ".user $@\n")) dienomem() ;
   auto_file(compiled, fn, satmp.s + base, satmp.len - base) ;
   satmp.len = base ;
   auto_rights(compiled, fn, 0755) ;
@@ -1309,7 +1327,7 @@ static inline void write_servicedirs (char const *compiled, s6rc_db_t const *db,
       memcpy(dstfn + clen + 14 + len, "finish", 7) ;
       if (ispipelined)
       {
-        write_exe_wrapper(compiled, dstfn + clen + 1, db, i, "finish", 1) ;
+        write_exe_wrapper(compiled, dstfn + clen + 1, db, i, "finish") ;
         memcpy(dstfn + clen + 20 + len, ".user", 6) ;
       }
       if (!filecopy_unsafe(srcfn, dstfn, 0755))
@@ -1323,7 +1341,7 @@ static inline void write_servicedirs (char const *compiled, s6rc_db_t const *db,
     memcpy(dstfn + clen + 14 + len, "run", 4) ;
     if (ispipelined)
     {
-      write_exe_wrapper(compiled, dstfn + clen + 1, db, i, "run", 0) ;
+      write_exe_wrapper(compiled, dstfn + clen + 1, db, i, "run") ;
       memcpy(dstfn + clen + 17 + len, ".user", 6) ;
     }
     if (!filecopy_unsafe(srcfn, dstfn, 0755))
@@ -1384,6 +1402,24 @@ static inline void write_servicedirs (char const *compiled, s6rc_db_t const *db,
     memcpy(dstfn + clen + 14 + len, "down-signal", 12) ;
     filecopy_unsafe(srcfn, dstfn, 0644) ;
 
+    memcpy(srcfn + srcdirlen + len + 2, "template", 9) ;
+    memcpy(dstfn + clen + 14 + len, "template", 9) ;
+    if (dircopy(compiled, srcfn, dstfn))
+    {
+      memcpy(dstfn + clen + 14 + len, "instances", 10) ;
+      if (mkdir(dstfn, 0755) == -1)
+      {
+        cleanup(compiled) ;
+        strerr_diefu2sys(111, "mkdir ", dstfn) ;
+      }
+      dstfn[clen + 22 + len] = 0 ;
+      if (mkdir(dstfn, 0755) == -1)
+      {
+        cleanup(compiled) ;
+        strerr_diefu2sys(111, "mkdir ", dstfn) ;
+      }
+    }
+
     memcpy(srcfn + srcdirlen + len + 2, "data", 5) ;
     memcpy(dstfn + clen + 14 + len, "data", 5) ;
     dircopy(compiled, srcfn, dstfn) ;
@@ -1482,7 +1518,6 @@ static inline void write_db (char const *compiled, s6rc_db_t const *db)
   if (buffer_putflush(&b, S6RC_DB_BANNER_END, S6RC_DB_BANNER_END_LEN) < 0)
     goto err ;
 
-  if (fsync(fd) < 0) goto err ;
   close(fd) ;
   return ;
  err:
@@ -1513,29 +1548,32 @@ static inline void write_compiled (
 
 int main (int argc, char const *const *argv)
 {
-  before_t before = BEFORE_ZERO ;
-  char const *compiled ;
-  char const *fdhuser = 0 ;
-  int blocking = 0 ;
-  PROG = "s6-rc-compile" ;
+  static gol_bool const rgolb[] =
   {
-    subgetopt l = SUBGETOPT_ZERO ;
-    for (;;)
-    {
-      int opt = subgetopt_r(argc, argv, "v:u:g:h:b", &l) ;
-      if (opt == -1) break ;
-      switch (opt)
-      {
-        case 'v' : if (!uint0_scan(l.arg, &verbosity)) dieusage() ; break ;
-        case 'h' : fdhuser = l.arg ; break ;
-        case 'b' : blocking = 1 ; break ;
-        default : dieusage() ;
-      }
-    }
-    argc -= l.ind ; argv += l.ind ;
-  }
+    { .so = 'b', .lo = "block", .clear = 0, .set = GOLB_BLOCK },
+  } ;
+  static gol_arg const rgola[] =
+  {
+    { .so = 'v', .lo = "verbosity", .i = GOLA_VERBOSITY },
+    { .so = 'h', .lo = "fdholder-user", .i = GOLA_FDHUSER },
+  } ;
+  before_t before = BEFORE_ZERO ;
+  uint64_t wgolb = 0 ;
+  char const *wgola[GOLA_N] = { 0 } ;
+  char const *compiled ;
+  unsigned int golc ;
+  PROG = "s6-rc-compile" ;
+
+  golc = GOL_main(argc, argv, rgolb, rgola, &wgolb, wgola) ;
+  argc -= golc ; argv += golc ;
   if (argc < 2) dieusage() ;
+  if (wgola[GOLA_VERBOSITY])
+  {
+    if (!uint0_scan(wgola[GOLA_VERBOSITY], &verbosity))
+      strerr_dief1x(100, "verbosity must be an unsigned integer") ;
+  }
   compiled = *argv++ ;
+
   before.specialdeps[0] = add_internal_longrun(&before, S6RC_ONESHOT_RUNNER) ;
   before.specialdeps[1] = add_internal_longrun(&before, S6RC_FDHOLDER) ;
   {
@@ -1582,9 +1620,9 @@ int main (int argc, char const *const *argv)
       db.deps = deps ;
       flatlist_services(&db, sarray) ;
       propagate_bundle_flags(&db, bundles, nbundles, bdeps) ;
-      write_compiled(compiled, &db, srcdirs, bundles, nbundles, bdeps, fdhuser, blocking) ;
+      write_compiled(compiled, &db, srcdirs, bundles, nbundles, bdeps, wgola[GOLA_FDHUSER], wgolb & GOLB_BLOCK) ;
     }
   }
 
-  return 0 ;
+  _exit(0) ;
 }
