@@ -32,8 +32,8 @@ int main (int argc, char const *const *argv)
   tain deadline ;
   stralloc sa = STRALLOC_ZERO ;
   size_t dirlen ;
-  char const *live = S6RC_LIVE_BASE ;
-  char const *compiled = S6RC_COMPILED_BASE ;
+  char const *live = S6RC_LIVEDIR ;
+  char const *compiled = S6RC_BOOTDB ;
   char const *prefix = "" ;
   int deref = 0 ;
   PROG = "s6-rc-init" ;
@@ -73,6 +73,8 @@ int main (int argc, char const *const *argv)
   tain_now_set_stopwatch_g() ;
   tain_add_g(&deadline, &deadline) ;
 
+  if (!s6rc_livedir_canon(&live))
+    strerr_diefu1sys(111, "canonicalize livedir") ;
   if (deref)
   {
     char *x = realpath(compiled, 0) ;
@@ -113,7 +115,7 @@ int main (int argc, char const *const *argv)
       strerr_diefu4sys(111, "recursively copy ", cfn, " to ", lfn) ;
     }
   }
-  if (!atomic_symlink(sa.s + dirlen, live, PROG))
+  if (!atomic_symlink4(sa.s + dirlen, live, 0, 0))
   {
     cleanup(&sa) ;
     strerr_diefu4sys(111, "symlink ", sa.s + dirlen, " to ", live) ;
